(* Talk-Patcher / Installer *)
open Util
open Diff

let always_yes = ref false 
let always_uninstall = ref false 

type tp_flag =
  Auto_Tra of string
| Menu_Style of string
| Ask_Every_Component 
| Always of tp_action list 

and tp_file = {
  backup    : string ;
  author    : string ; 
  flags     : tp_flag list ; 
  languages : tp_lang list ;
  modules   : tp_mod list ;
  allow_missing : string list ; 
} 

and tp_lang = {
  lang_name : string ;
  lang_dir_name : string ;
  lang_tra_files : string list ;
} 

and tp_mod = {
  mod_name : Dlg.tlk_string ;
  mod_parts : tp_action list;
  mod_flags : tp_mod_flag list ; 
} 

and tp_mod_flag = 
  TPM_Deprecated of Dlg.tlk_string (* should be uninstalled when encountered *) 
| TPM_RequireComponent of string * int * Dlg.tlk_string  

and tp_copy_args = {
  copy_get_existing   : bool ;  (* get from biffs? *)
  copy_use_regexp     : bool ;
  copy_use_glob       : bool ;
  copy_file_list      : ( string * string ) list ; (* (source,dest) list *)
  copy_patch_list     : tp_patch list ;
  copy_constraint_list: tp_constraint list ;
  copy_backup         : bool ; (* normally TRUE *)
} 

and tp_action =
  | TP_Copy   of tp_copy_args 
  | TP_CopyRandom of string list * tp_patch list 
  | TP_RandomSeed of int 
  | TP_Compile of (string list) * (string list) (* DLG, TRA *)
  | TP_Mkdir of string list 
  | TP_Require_File of string * (Dlg.tlk_string)
  | TP_Forbid_File of string * (Dlg.tlk_string)
  | TP_Append of string * string * (tp_constraint list)
  | TP_Append_Col of string * (string list) * (tp_constraint list)
  | TP_Set_Col of string * (string list) * int 
  | TP_Extend_Top of bool * string * string * (tp_patch list) * (string list)
  | TP_Extend_Bottom of bool * string * string * (tp_patch list) 
      * (string list) 
  | TP_At_Exit of string 
  | TP_At_Interactive_Exit of string 
  | TP_At_Uninstall of string 
  | TP_At_Interactive_Uninstall of string 
  | TP_Add_Kit of tp_add_kit
  | TP_Add_Music of tp_add_music 
  | TP_String_Set of string * Dlg.tlk_string 
  | TP_Fail of Dlg.tlk_string 
  | TP_Print of Dlg.tlk_string 
  | TP_If of predicate * (tp_action list) * (tp_action list)  
  | TP_Uninstall_Now of string * int 

and predicate = 
  | Pred_And of predicate * predicate
  | Pred_Or of predicate * predicate
  | Pred_Not of predicate
  | Pred_File_MD5 of string * string
  | Pred_File_Exists of string
  | Pred_File_Exists_In_Game of string
  | Pred_File_Size of string * int
  | Pred_File_Contains of string * string

and tp_add_kit = {
  kit_name : string ;
  clasweap : string ;
  weapprof : string ;
  abclasrq : string ;
  abclsmod : string ;
  abdcdsrq : string ;
  abdcscrq : string ;
  dualclas : string ;
  alignmnt : string ;
  ability_file : string ; 
  include_in : string ;
  lower : Dlg.tlk_string ;
  mixed : Dlg.tlk_string ;
  help : Dlg.tlk_string ;
  unused_class : string ;
  tob_start : string list ;
  tob_abbr : string ; 
} 

and tp_add_music = {
  music_name : string ;
  music_file : string ; 
} 

and tp_patch = 
  | TP_PatchStrRef of tp_patchexp * Dlg.tlk_string (* offset + text *)
  | TP_PatchString of string * Dlg.tlk_string (* regexp + text *) 
  | TP_PatchStringTextually of string * string (* regexp + text *) 
  | TP_PatchReplaceBCSBlock of string * string (* old + new *) 
  | TP_PatchReplaceBCSBlockRE of string * string (* old + new *) 
  | TP_PatchApplyBCSPatch of string (* patch *)
  | TP_PatchByte of tp_patchexp * tp_patchexp
  | TP_PatchShort of tp_patchexp * tp_patchexp
  | TP_PatchLong of tp_patchexp * tp_patchexp
  | TP_PatchReadByte of tp_patchexp * string
  | TP_PatchReadShort of tp_patchexp * string
  | TP_PatchReadLong of tp_patchexp * string
  | TP_PatchASCII of tp_patchexp * string
  | TP_PatchInsertBytes of tp_patchexp * tp_patchexp 
  | TP_PatchDeleteBytes of tp_patchexp * tp_patchexp 
  | TP_PatchSet of string * tp_patchexp 
  | TP_PatchWhile of tp_patchexp * (tp_patch list)
  | TP_Patch2DA of tp_patchexp * tp_patchexp * tp_patchexp * string
  (* | TP_PatchIf of tp_patchexp * (tp_patch list) * (tp_patch list) *)

and tp_constraint = 
  | TP_Contains of string
  | TP_NotContains of string 
  | TP_IfSizeIs of int 
  | TP_Eval of tp_patchexp 

and tp_patchexp =
  | PE_String of string
  | PE_Not of tp_patchexp
  | PE_Add of tp_patchexp * tp_patchexp 
  | PE_Sub of tp_patchexp * tp_patchexp 
  | PE_Mul of tp_patchexp * tp_patchexp 
  | PE_Div of tp_patchexp * tp_patchexp 
  | PE_Equal of tp_patchexp * tp_patchexp 
  | PE_And of tp_patchexp * tp_patchexp 
  | PE_Or of tp_patchexp * tp_patchexp 
  | PE_GT of tp_patchexp * tp_patchexp 
  | PE_GTE of tp_patchexp * tp_patchexp 
  | PE_LT of tp_patchexp * tp_patchexp 
  | PE_LTE of tp_patchexp * tp_patchexp 
  | PE_If of tp_patchexp * tp_patchexp * tp_patchexp 

let strings_to_print_at_exit = ref [] 

let rec get_menu_style fl = match fl with
    [] -> 0
  | Menu_Style(i) :: tl -> int_of_string i
  | hd :: tl -> get_menu_style tl 

(************************************************************************
 * Evaluate a TP2 predicate
 ************************************************************************)
let rec eval game p = match p with 
  | Pred_And(p1,p2) -> (eval game p1) && (eval game p2) 
  | Pred_Or(p1,p2) -> (eval game p1) || (eval game p2)
  | Pred_Not(p1) -> not (eval game p1) 
  | Pred_File_MD5(f,s) -> 
      if file_size f > 0 then begin
        let digest = Digest.file f in
        let hex = Digest.to_hex digest in
        log_only "File [%s] has MD5 checksum [%s]\n" f hex ; 
        (String.uppercase hex) = (String.uppercase s)
      end else begin
        log_only "WARNING: File [%s] not found for MD5 checksum.\n" f ;
        false
      end
  | Pred_File_Exists(f) -> file_size f > 0
  | Pred_File_Exists_In_Game(f) -> 
      (try 
        let a,b = split f in
        let buff,path = Load.load_resource game true a b in
        true
      with _ -> false )
  | Pred_File_Size(f,s) -> file_size f = s
  | Pred_File_Contains(f,r) -> 
      let buf = load_file f in
      let regexp = Str.regexp_case_fold r in 
      try 
        let _ = Str.search_forward regexp buf 0 in 
        true
      with Not_found -> false

(************************************************************************
 * Uninstall a TP2 component
 ************************************************************************)
let uninstall_tp2_component tp2 tp_file i interactive =
  try 
    let result = tp2 in
    let d = result.backup ^ "/" ^ (string_of_int i) in 
    let u_filename = (Printf.sprintf "%s/UNINSTALL.%d" d i) in 
    let buff = load_file u_filename in 
    let file_list = Str.split (Str.regexp "\n") buff in
    log_and_print "Will uninstall %d files for [%s] component %d.\n" 
      (List.length file_list) tp_file i; 
    List.iter (fun override_filename -> 
      let base = Filename.basename override_filename in 
      let backup_filename = d ^ "/" ^ base in 
      (* did we save a backup copy? *) 
      if file_size backup_filename > 0 then begin
        (* copy it from Backup to Override *) 
        log_or_print "  Restoring backed-up [%s]\n" backup_filename ; 
        let buff = load_file backup_filename in
        let restore_chn = open_out_bin override_filename in
        output_string restore_chn buff ;
        close_out restore_chn ;
        Unix.unlink backup_filename 
      end else begin
        (* no backup copy, so just delete it from override *) 
        log_or_print "  Deleting [%s] (to restore original game condition)\n" 
          override_filename ; 
        try 
          Unix.unlink override_filename  
        with e ->
          log_or_print "Unable to Unlink [%s]: %s"
            override_filename (Printexc.to_string e)
      end 
    ) file_list  ;
    Unix.unlink u_filename ;
    let m = List.nth result.modules i in 
    List.iter (fun a -> match a with
      | TP_At_Uninstall(str) -> 
        let str = Arch.handle_view_command str in 
        ignore (Unix.system(Arch.slash_to_backslash str))
      | TP_At_Interactive_Uninstall(str) -> 
        let str = Arch.handle_view_command str in 
        if interactive then ignore(Unix.system(Arch.slash_to_backslash str))
      | _ -> () 
    ) m.mod_parts ; 
    log_and_print "Uninstalled    %d files for [%s] component %d.\n" 
      (List.length file_list) tp_file i; 
  with e ->
    log_and_print "Error Uninstalling [%s] component %d:\n%s\n"
      tp_file i (Printexc.to_string e)

(************************************************************************
 * Determine what has been installed. 
 ************************************************************************)
type status = Installed | Temporarily_Uninstalled | Permanently_Uninstalled
type installed_mods = (string * int * int * (string option) * status) list
  (* module : language : component : status *) 
let log_name = "WeiDU.log"
let the_log = ref [] 

let str_of_str_opt sopt = match sopt with
    Some(str) -> "~" ^ str ^ "~"
  | None -> ""

let print_log () = 
  List.iter (fun (n,i1,i2,sopt,st) ->
    log_or_print "%s %d %d %s %s\n" n i1 i2 (match st with
      Installed -> "Installed"
    | Temporarily_Uninstalled -> "Temporarily_Uninstalled"
    | Permanently_Uninstalled -> "Permanently_Uninstalled")
    (str_of_str_opt sopt) 
  ) !the_log

let save_log () =
  let out = open_out log_name in
  Printf.fprintf out "// Log of Currently Installed WeiDU Mods\n" ;
  Printf.fprintf out "// The top of the file is the 'oldest' mod\n" ; 
  Printf.fprintf out "// ~TP2_File~ #language_number #component_number\n" ;
  List.iter (fun (a,b,c,sopt,d) -> match d with
    Installed -> Printf.fprintf out "~%s~ #%d #%d %s\n" (String.uppercase a) b c
      (str_of_str_opt sopt) 
  | Temporarily_Uninstalled -> 
    log_and_print "Internal Error: saving a log with temporarily uninstalled module [%s] %d %d %s\n" a b c (str_of_str_opt sopt)
  | Permanently_Uninstalled ->
      Printf.fprintf out "// Recently Uninstalled: ~%s~ #%d #%d %s\n"
      (String.uppercase a) b c
      (str_of_str_opt sopt)
  ) !the_log ; 
  close_out out 

let log_match a b = 
  Filename.basename (String.uppercase a) = Filename.basename (String.uppercase b) 

let already_installed tp2 i =
  let rec is_installed lst = match lst with
  | [] -> false
  | (a,b,c,sopt,d) :: tl when log_match a tp2 
        && c = i && d <> Permanently_Uninstalled -> true
  | hd :: tl -> is_installed tl 
  in is_installed !the_log 

let temporarily_uninstalled tp2 i =
  let rec is_installed lst = match lst with
  | [] -> false
  | (a,b,c,sopt,d) :: tl when log_match a tp2
        && c = i && d = Temporarily_Uninstalled -> true
  | hd :: tl -> is_installed tl 
  in is_installed !the_log 

let temp_to_perm_uninstalled tp2 i = 
  let rec is_installed lst = match lst with
  | [] -> []
  | (a,b,c,sopt,d) :: tl when log_match a tp2 
        && c = i && d = Temporarily_Uninstalled ->
          (a,b,c,sopt,Permanently_Uninstalled) :: tl 
  | hd :: tl -> hd :: (is_installed tl)
  in the_log := is_installed !the_log 

(************************************************************************
 * Do everything necessary to uninstall the given tp2 component. This
 * may require temporarily uninstalling stuff that has been installed
 * since our target was installed. 
 ************************************************************************)
let rec find_best_file lst =
  match lst with
    hd :: tl -> if file_size hd > 0 then hd 
                else if file_size (Filename.basename hd) > 0 then
                  (Filename.basename hd)
                else find_best_file tl
  | [] -> failwith "TP2 not found!"  

let uninstall handle_tp2_filename tp2 i interactive =
  log_or_print "uninstall: %s %d\n" tp2 i ; 
  let worked = ref true in 
  if not (already_installed tp2 i) then begin
    log_and_print "Internal Error: trying to uninstall non-installed mod component [%s] %d\n" tp2 i ;
    false 
  end else if (temporarily_uninstalled tp2 i) then begin
    log_or_print "uninstall: %s %d already temporarily uninstalled\n" tp2 i ;
    temp_to_perm_uninstalled tp2 i ; 
    true 
  end else begin
    let rec prepare lst = match lst with
    | [] -> [] (* end of the line *)

    (* this is the entry in the list *)
    | (a,b,c,sopt,d) :: tl when log_match a tp2 && c = i -> 
      begin match d with
      | Permanently_Uninstalled -> (* some sort of error *) 
                  log_and_print "Internal Error: mod component [%s] %d already uninstalled\n" tp2 i ; (a,b,c,sopt,Permanently_Uninstalled) :: tl
      | Temporarily_Uninstalled -> (* we just won't restore it! *)
                      (a,b,c,sopt,Permanently_Uninstalled) :: tl
      | Installed -> 
          begin 
          try 
            let best = find_best_file [a ; tp2] in 
            uninstall_tp2_component (handle_tp2_filename best) a c interactive ;
            (a,b,c,sopt,Permanently_Uninstalled) :: tl
          with _ ->
            log_and_print "ERROR: This Mod is too old (or too new) to uninstall that component for you.\nUpgrade to the newest versions of this mod and that one and try again.\n" ;
            worked := false ; 
            lst 
          end 
      end

    | (a,b,c,sopt,d) :: tl -> 
      begin match d with
      | Permanently_Uninstalled    
      | Temporarily_Uninstalled -> (* keep going *)
                      (a,b,c,sopt,d) :: (prepare tl)
      | Installed -> 
          log_or_print "We must temporarily uninstall [%s] component %d\n" 
            a c ; 
          begin
          try
            let best = find_best_file [ a] in 
            uninstall_tp2_component (handle_tp2_filename best) a c false ; 
            (* take away for now *)
            (a,b,c,sopt,Temporarily_Uninstalled) :: (prepare tl)
          with e -> 
            log_and_print "ERROR: This Mod is too old (or too new) to uninstall that component for you.\nUpgrade to the newest versions of this mod and that one and try again.\n[%s]\n" 
            (Printexc.to_string e);
            worked := false ; 
            lst 
          end
      end
    in 
    let new_log = List.rev (prepare (List.rev !the_log)) in
    the_log := new_log ;
    !worked
  end

let body_of_script buff = 
  if buff = "" then "" else 
  try 
    let first_nl = String.index buff '\n' in
    let last_nl = (String.rindex buff 'S') - 1 in 
    if first_nl = last_nl then 
      "" 
    else 
      let length = (last_nl - first_nl) - 1 in 
      String.sub buff (first_nl + 1) length 
  with e -> 
    log_and_print "ERROR: not a BCS script\n" ;
    failwith "not a BCS script" 

type default_action = TP_Install | TP_Uninstall | TP_Skip | TP_Ask 

let is_true i = (Int32.compare i (Int32.zero)) <> 0 

let rec eval_pe p = 
  match p with
  | PE_String(s) -> 
      (try Int32.of_string (Var.substitute s)
       with e -> log_and_print "ERROR: cannot convert ~%s~ (or ~%s~) to an integer\n" s (Var.substitute s) ; raise e)
  | PE_Add(a,b) -> let a = eval_pe a in let b = eval_pe b in Int32.add a b 
  | PE_Sub(a,b) -> let a = eval_pe a in let b = eval_pe b in Int32.sub a b 
  | PE_Mul(a,b) -> let a = eval_pe a in let b = eval_pe b in Int32.mul a b 
  | PE_Div(a,b) -> let a = eval_pe a in let b = eval_pe b in 
    (try Int32.div a b with Division_by_zero -> Int32.zero)
  | PE_Not(a) ->   if is_true (eval_pe a) 
                   then Int32.zero else Int32.one
  | PE_And(a,b) -> if is_true (eval_pe a) && is_true (eval_pe b) 
                  then Int32.one else Int32.zero 
  | PE_Or(a,b) -> if is_true (eval_pe a) || is_true (eval_pe b) 
                  then Int32.one else Int32.zero 
  | PE_Equal(a,b) -> if Int32.compare (eval_pe a) (eval_pe b) = 0 
                  then Int32.one else Int32.zero 
  | PE_GT(a,b) -> if Int32.compare (eval_pe a) (eval_pe b) > 0 
                  then Int32.one else Int32.zero 
  | PE_GTE(a,b) -> if Int32.compare (eval_pe a) (eval_pe b) >= 0
                  then Int32.one else Int32.zero 
  | PE_LT(a,b) -> if Int32.compare (eval_pe a) (eval_pe b) < 0 
                  then Int32.one else Int32.zero 
  | PE_LTE(a,b) -> if Int32.compare (eval_pe a) (eval_pe b) <= 0 
                  then Int32.one else Int32.zero 
  | PE_If(p,t,e) -> if is_true (eval_pe p) then eval_pe t else eval_pe e



(************************************************************************
 * change location "where" in "buff" to point to str-ref "what"
 ************************************************************************)
let process_patch1 game buff p = match p with 
  | TP_PatchReadByte(where,name) ->
      let where = Int32.to_int (eval_pe where) in 
      let value = byte_of_str_off buff where in
      Var.add_var name (Printf.sprintf "%d" value)
  | TP_PatchReadShort(where,name) ->
      let where = Int32.to_int (eval_pe where) in 
      let value = short_of_str_off buff where in
      Var.add_var name (Printf.sprintf "%d" value) 
  | TP_PatchReadLong(where,name) ->
      let where = Int32.to_int (eval_pe where) in 
      let value = int32_of_str_off buff where in
      Var.add_var name (Printf.sprintf "%ld" value) 
  | _ -> () 

let rec process_patch2 game buff p =
  match p with
    TP_PatchStrRef(where,what) -> 
      let where = Int32.to_int (eval_pe where) in 
      let new_index = match Dc.resolve_tlk_string game what with
        Dlg.TLK_Index(i) -> i
      | _ -> log_and_print "ERROR: cannot resolve SAY patch\n" ; failwith "resolve" 
      in 
      let new_string = str_of_int new_index in 
      String.blit new_string 0 buff where 4 ;
      buff

  | TP_PatchReplaceBCSBlock(old_file, new_file) -> begin
      let old_file_buff = load_file old_file in  
      let string_to_find = body_of_script old_file_buff in
      let new_file_buff = load_file new_file in  
      let string_to_sub_in = body_of_script new_file_buff in
      let my_regexp = Str.regexp_string string_to_find in
      try 
        let _ = Str.search_forward my_regexp buff 0 in
        Str.global_replace my_regexp string_to_sub_in buff 
      with Not_found -> 
        log_and_print "WARNING: cannot find block matching [%s]\n"
          old_file ;
        buff
      end 

  | TP_PatchReplaceBCSBlockRE(old_file, new_file) -> begin
      let string_to_find = load_file old_file in
      let string_to_sub_in = load_file new_file in
      let my_regexp = Str.regexp string_to_find in
      try 
        let _ = Str.search_forward my_regexp buff 0 in
        Str.global_replace my_regexp string_to_sub_in buff 
      with Not_found -> 
        log_and_print "WARNING: cannot find block matching [%s]\n"
          old_file ;
        buff
      end 

  | TP_PatchApplyBCSPatch(patch_file) -> begin
      let patch_buff = load_file patch_file in
      try
        let new_buff, bad_chunks, app_chunks = Diff.do_patch buff patch_buff false in begin
          if ( bad_chunks > 0 ) then begin
            log_and_print "ERROR: Cannot apply patch %s (%d bad chunks).\n" patch_file bad_chunks ;
            failwith "Cannot Apply Patch"
          end ;
          if ( app_chunks > 0 ) then begin
            log_and_print "WARNING: %d chunks in patch file %s already applied.\n" app_chunks patch_file 
          end ;
          new_buff
        end
      with Not_found ->
        log_and_print "Error: applying patch %s failed.\n" patch_file ;
        buff
  end

  | TP_PatchInsertBytes(where,how_many) ->
      let how_many = Int32.to_int (eval_pe how_many) in 
      let blanks = String.make how_many ('\000') in
      let where = Int32.to_int (eval_pe where) in 
      let before = Str.string_before buff where in
      let after = Str.string_after buff where in
      before ^ blanks ^ after

  | TP_PatchDeleteBytes(where,how_many) ->
      let where = Int32.to_int (eval_pe where) in 
      let before = Str.string_before buff where in
      let how_many = Int32.to_int (eval_pe how_many) in 
      let after = Str.string_after buff (where + how_many) in
      before ^ after

  | TP_PatchSet(name,value) -> 
      let value = (eval_pe value) in
      Var.add_var name (Int32.to_string value) ;
      buff

  | TP_PatchWhile(pred,pl) -> 
      let b = ref buff in
      while is_true (eval_pe pred) do
        b := List.fold_left (fun acc elt -> 
            process_patch1 game acc elt; 
            process_patch2 game acc elt) !b pl 
      done ;
      !b

  | TP_Patch2DA(row,col,req_col,value) ->
      let row = Int32.to_int (eval_pe row) in
      let col = Int32.to_int (eval_pe col) in
      let req_col = Int32.to_int (eval_pe req_col) in 
      let lines = Str.split (Str.regexp "[\r\n]+") buff in
      let max = ref (String.length value) in 
      let entries = List.map (fun line -> 
        let entry_list = Str.split (Str.regexp "[ \t]+") line in
        List.iter (fun e ->
          let len = String.length e in
          if len > !max then max := len
        ) entry_list ; 
        Array.of_list entry_list) lines in
      let rec process line_list lines_left = 
        match line_list, lines_left with
          [], 0 -> ()
        | [], _ -> 
          begin
            log_and_print "ERROR: Cannot find %d rows with at least %d columns." row req_col ;
            failwith "Cannot Set 2DA Entry" 
          end
        | hd :: tl, i  when Array.length hd < req_col -> process tl i 
        | hd :: tl, 0 -> hd.(col) <- value 
        | hd :: tl, i -> process tl (i - 1)
      in
      process entries row ;
      List.fold_left (fun line_acc entry_array -> 
        line_acc ^ (Array.fold_left (fun acc elt ->
          Printf.sprintf "%s%-.*s\t" acc !max elt) "" entry_array) ^ "\n") "" entries

  | TP_PatchStringTextually(find,what) -> 
      let my_regexp = Str.regexp_string find in
      let what = Var.substitute what in 
      Str.global_replace my_regexp what buff 

  | TP_PatchString(find,what) -> 
      let my_regexp = Str.regexp_string find in
      let new_index = match Dc.resolve_tlk_string game what with
        Dlg.TLK_Index(i) -> i
      | _ -> log_and_print "ERROR: cannot resolve REPLACE patch\n" ; failwith "resolve" 
      in 
      let new_string = Printf.sprintf "%d" new_index in 
      Str.global_replace my_regexp new_string buff 

  | TP_PatchByte(where,what) -> 
      let where = Int32.to_int (eval_pe where) in 
      let what = Int32.to_int (eval_pe what) in 
      let str = String.make 1 (Char.chr what) in
      String.blit str 0 buff where 1 ;
      buff
  | TP_PatchShort(where,what) -> 
      let where = Int32.to_int (eval_pe where) in 
      let what = Int32.to_int (eval_pe what) in 
      let str = str_of_short what in 
      String.blit str 0 buff where 2 ;
      buff
  | TP_PatchLong(where,what) -> 
      let where = Int32.to_int (eval_pe where) in 
      let what = eval_pe what in 
      let str = str_of_int32 what in 
      String.blit str 0 buff where 4 ;
      buff
  | TP_PatchASCII(where,what) -> 
      let where = Int32.to_int (eval_pe where) in 
      String.blit what 0 buff where (String.length what) ;
      buff

  | _ -> buff 

let handle_tp 
      game 
      handle_tp2_filename
      handle_tra_filename 
      handle_d_filename
      compile_baf_filename
      handle_script_buffer
      emit_dlg_files
      this_tp2_filename
      tp 
= begin
  Load.allow_missing := tp.allow_missing ; 
  let interactive = ref true in 
  let our_lang = ref None in
  let our_lang_index = ref 0 in 
  (* pick your language *)
  begin 
    match tp.languages with
      [] -> ()
    | [l] -> our_lang := Some(l) 
    | _ -> 
      let arr = Array.of_list tp.languages in
      let answer = ref (None) in 
      let answer_index = ref 0 in 
      while !answer = None do  
        log_and_print "\nChoose your language:\n" ;
        Array.iteri (fun i l -> 
          log_and_print "%2d [%s]\n" i l.lang_name) arr ;
        try 
          let i = read_int () in
          if i >= 0 && i < Array.length arr then begin
            answer := Some(arr.(i)) ;
            answer_index := i 
          end 
        with _ -> () 
      done ;
      our_lang := !answer ;
      our_lang_index := !answer_index 
  end ;
      (* 2da helper functions *)
      let split_apart str =
        Str.split (Str.regexp "[ \t]+") str 
      in 
      let get_next_col_number file =
        let (a,b) = split file in 
        let buff,path = Load.load_resource game true a b in 
        try 
          let lst = Str.split (Str.regexp "[\r\n]+") buff in
          let elt = List.nth lst 5 in
          let cols = split_apart elt in 
          let num_cols = List.length cols in 
          let last_col = num_cols - 2 in 
          last_col + 1
        with e -> 
          log_and_print "ERROR: cannot find col numbers in %s\n" file ;
          raise e
      in 
      let get_next_line_number file =
        let (a,b) = split file in 
        let buff,path = Load.load_resource game true a b in 
        try 
          let idx = Str.search_backward (Str.regexp "[\r\n][0-9]") buff 
            ((String.length buff) - 1) in 
          let minibuff = Str.string_after buff (idx+1) in
          let lst = split_apart minibuff in
          let elt = List.hd lst in
          let last_number = int_of_string elt in
          last_number + 1
        with e -> 
          log_and_print "ERROR: cannot find line numbers in %s\n" file  ;
          raise e
      in 
      (* 2da helper functions end *)
  let rec process_action tp a = match a with
    | TP_Require_File(file,error_msg) ->
        log_and_print "Checking for required files ...\n" ;
        let size = file_size file in 
        if size > 0 then begin
          log_or_print "[%s] %d bytes" file size 
        end else begin
          log_and_print "[%s] not found\n" file ;
          log_and_print "\n%s\n" (Dc.single_string_of_tlk_string game
          error_msg) ;
          failwith file
        end

    | TP_Print(msg) -> 
        let str = Dc.single_string_of_tlk_string game msg in 
        let str = Var.substitute str in 
        log_and_print "\n%s\n" str 

    | TP_Fail(msg) -> 
        let str = Dc.single_string_of_tlk_string game msg in 
        log_and_print "FAILURE:\n%s\n" str ;
        failwith str 

    | TP_If(p,al1,al2) -> 
        let res = eval game p in
        log_or_print "IF evaluates to %b\n" res ;
        if res then begin
          List.iter (process_action tp) al1
        end else begin
          List.iter (process_action tp) al2
        end 

    | TP_Forbid_File(file,error_msg) ->
        log_and_print "Checking for forbidden files ...\n" ;
        let size = file_size file in 
        if size > 0 then begin
          log_and_print "[%s] found: %d bytes\n" file size ;
          log_and_print "\n%s\n" (Dc.single_string_of_tlk_string game
          error_msg) ;
          failwith file
        end else begin
          log_or_print "[%s] not found (as desired)\n" file ;
        end

    | TP_Uninstall_Now(name,comp) -> 
        if already_installed name comp then begin
          if uninstall handle_tp2_filename name comp false then
            () 
          else 
            failwith 
              (Printf.sprintf "unable to uninstall %s component #%d"
                name comp) 
        end else
          log_or_print "%s component #%d not present, good.\n"
            name comp 

    | TP_RandomSeed(i) -> Random.init i

    | TP_CopyRandom(slist,plist) ->
        log_and_print "Randomizing %d file(s) ...\n" (List.length slist) ; 
        let blist = List.map (fun s -> 
          let a,b = split s in 
          let buff,path = Load.load_resource game true a b in
          (s,buff)
        ) slist in
        let dlist = List.map (fun s -> "override/" ^ s) slist in
        let rand_dlist = List.sort 
          (fun a b -> if Random.bool () then 1 else -1) dlist in
        List.iter2 (fun (s,buff) d ->
          List.iter (fun p -> process_patch1 game buff p) plist ; 
          let result_buff = 
            List.fold_left (fun acc elt -> 
              try process_patch2 game acc elt
              with e -> 
                log_and_print "ERROR: [%s] -> [%s] Patching Failed (COPY_RANDOM) (%s)\n" 
                  s d (Printexc.to_string e); raise e) 
              buff plist 
          in 
            Stats.time "saving files" (fun () -> 
            let out = open_for_writing d true in
            output_string out result_buff ;
            close_out out ;
            log_only "Copied [%s] to [%s]\n" s d) ()
        ) blist rand_dlist 

    | TP_Copy(copy_args) -> 
        let get_existing = copy_args.copy_get_existing in
        let use_reg = copy_args.copy_use_regexp in
        let use_glob = copy_args.copy_use_glob in 
        let slist = copy_args.copy_file_list in
        let plist = copy_args.copy_patch_list in 
        let clist = copy_args.copy_constraint_list in
        let make_a_backup = copy_args.copy_backup in 

        log_and_print "Copying %d file(s) ...\n" (List.length slist) ; 

        let slist = 
          if use_glob then begin
            let res = ref [] in 
            List.iter (fun (s,p) -> 
              let myfun (glob_s : string) = begin
                log_only "Callback from Arch.glob: %s\n" glob_s; 
                res := (glob_s, p) :: !res
              end in
              log_only "Calling Arch.glob: %s\n" s; 
              let res = (Arch.glob s myfun) in 
              log_only "Arch.glob returned.\n"; 
              res
            ) slist ;
            !res
          end else if get_existing = false || use_reg = false then
            slist
          else begin
            let files_in_chitin = Key.list_of_key_resources game.Load.key in 
            let new_list = List.map (fun (s,p) ->
              let regexp = Str.regexp_case_fold s in 
              let matches = ref [] in 
              List.iter (fun possible ->
                if Str.string_match regexp possible 0 then begin
                  matches := (possible, p ^ "/" ^ possible) :: !matches
                end 
              ) files_in_chitin ;
              if (!matches = []) then
                [(s,p)]
              else
                !matches
            ) slist in 
            List.flatten new_list 
          end
        in 

        let copy_one_file src dest = begin
          let src_dir = Filename.dirname src in
          Var.add_var "SOURCE_DIRECTORY" src_dir ; 
          let dest = Var.substitute_user_vars_only dest in 
          let buff = 
            if not get_existing then 
              load_file src 
            else
              let a,b = split src in 
              let buff,path = Load.load_resource game true a b in
              buff
          in
          List.iter (fun p -> process_patch1 game buff p) plist ; 
          let ok_to_copy = List.fold_left (fun acc elt ->
            acc && match elt with
            | TP_Contains(s) -> begin
                let my_regexp = Str.regexp_string s in
                try let _ = Str.search_forward my_regexp buff 0 in
                  true
                with _ -> 
                  log_only "Not copying [%s] to [%s] because it does NOT contain [%s]\n" src dest s ; 
                  false
                end
            | TP_NotContains(s) -> begin
                let my_regexp = Str.regexp_string s in
                try let _ = Str.search_forward my_regexp buff 0 in
                  log_only "Not copying [%s] to [%s] because it DOES contain [%s]\n" src dest s ; 
                  false
                with _ -> 
                  true
                end
            | TP_IfSizeIs(size) -> 
                if String.length buff = size then 
                  true
                else begin
                  log_only "Not copying [%s] to [%s] because size is %d, NOT %d\n" src dest (String.length buff) size ; 
                  false
                end
            | TP_Eval(pe) -> 
                let v = eval_pe pe in 
                if v = Int32.zero then begin
                  log_only "Not copying [%s] to [%s] because condition evaluates to %ld\n" src dest v ; false
                end else true 
          ) true clist in 
          if ok_to_copy then begin 
            let result_buff = 
              List.fold_left (fun acc elt -> 
                try process_patch2 game acc elt
                with e -> 
                  log_and_print "ERROR: [%s] -> [%s] Patching Failed (COPY) (%s)\n" 
                    src dest (Printexc.to_string e); raise e) 
                buff plist 
            in 
            let dest = 
              if is_directory dest then
                dest ^ "/" ^ (Filename.basename src)
              else 
                dest
            in 
            Stats.time "saving files" (fun () -> 
            let out = 
              try open_for_writing_internal make_a_backup dest true 
              with e -> 
                log_and_print "ERROR: COPY ~%s~ ~%s~ FAILED: cannot open target\n" src dest ; raise e
            in
            output_string out result_buff ;
            close_out out ;
            if make_a_backup then 
              log_only "Copied [%s] to [%s]\n" src dest
            else 
              log_only "Copied [%s] to [%s] (NO BACKUP MADE!)\n" src dest) ()
          end
        end in 
        List.iter (fun (src,dest) -> 
          if is_directory src then begin
            try
              let dh = Unix.opendir src in 
              while true do
                let base = Unix.readdir dh in 
                let source = src ^ "/" ^ base in 
                if not (is_directory source) then
                  copy_one_file source (dest ^ "/" ^ base)
              done 
            with _ -> () 
          end else copy_one_file src dest 
        ) slist 

    | TP_Add_Music(m) -> begin
      log_and_print "Adding %s Music ...\n" m.music_name; 
      let this_music_number = get_next_line_number "SONGLIST.2DA" in 
      let music_base_name = Filename.basename m.music_file in 
      let str_to_append = Printf.sprintf "%d %s %s"
        this_music_number m.music_name music_base_name in 

      let a1 = TP_Append("SONGLIST.2DA",str_to_append,[]) in

      let dest_music_file = "music/" ^ music_base_name in 
      let a2 = TP_Copy(
        { copy_get_existing = false;
          copy_use_regexp = false;
          copy_use_glob = false;
          copy_file_list = [(m.music_file,dest_music_file)] ;
          copy_patch_list = [] ;
          copy_constraint_list = [] ;
          copy_backup = true; } ) in 
      let action_list = [ a1 ; a2 ] in 
      List.iter (process_action tp) action_list ; 
      Var.add_var (m.music_name) (string_of_int this_music_number) ;
      log_and_print "Added %s Music\n" m.music_name; 
    end

    | TP_Add_Kit(k) -> begin
      log_and_print "Adding %s Kit ...\n" k.kit_name; 

      if eval game 
        (Pred_And(
        (Pred_File_Exists("override/kitlist.2da")),
        (Pred_File_Contains("override/kitlist.2da",k.kit_name)))) then begin
        log_and_print "\n\nERROR: Kit [%s] already present! Skipping!\n\n"
        k.kit_name 
      end else begin 
        let a1 = TP_Append("CLASWEAP.2DA",k.clasweap,[]) in
        let a2 = TP_Append_Col("WEAPPROF.2DA",
          ("" :: "" :: (split_apart k.weapprof)),[]) in
        let a3 = TP_Append("ABCLASRQ.2DA",k.abclasrq,[]) in
        let a4 = TP_Append("ABDCDSRQ.2DA",k.abdcdsrq,[]) in
        let a5 = TP_Append("ABDCSCRQ.2DA",k.abdcscrq,[]) in
        let a_e1 = TP_Append("ABCLSMOD.2DA",k.abclsmod,[]) in
        let a_e2 = TP_Append("DUALCLAS.2DA",k.dualclas,[]) in
        let a6 = TP_Append("ALIGNMNT.2DA",k.alignmnt,[]) in
        let abil_file = String.uppercase (Filename.basename k.ability_file) in 
        let abil_file_no_ext = Filename.chop_extension abil_file in 
        let dest_abil_file = "override/" ^ abil_file in 
        let a7 = TP_Copy(
          { copy_get_existing = false ;
            copy_use_regexp = false;
            copy_use_glob = false;
            copy_file_list = [(k.ability_file,dest_abil_file)] ;
            copy_patch_list = [] ;
            copy_constraint_list = [] ;
            copy_backup = true ; } ) in 
        let include_list = split_apart k.include_in in
        let lower_index = match Dc.resolve_tlk_string game k.lower with
          Dlg.TLK_Index(i) -> i
        | _ -> log_and_print "ERROR: cannot resolve KIT lower string\n" ; 
                failwith "resolve" 
        in 
        let mixed_index = match Dc.resolve_tlk_string game k.mixed with
          Dlg.TLK_Index(i) -> i
        | _ -> log_and_print "ERROR: cannot resolve KIT mixed string\n" ; 
                failwith "resolve" 
        in 
        let help_index = match Dc.resolve_tlk_string game k.help with
          Dlg.TLK_Index(i) -> i
        | _ -> log_and_print "ERROR: cannot resolve KIT help string\n" ; 
                failwith "resolve" 
        in 
        let this_kit_number = get_next_line_number "KITLIST.2DA" in
        let this_kit_prof_number = get_next_col_number "WEAPPROF.2DA" in 
        let append_to_kitlist = Printf.sprintf 
          "%d  %s %d %d %d %s %d %s" 
          this_kit_number k.kit_name 
          lower_index mixed_index help_index
          abil_file_no_ext this_kit_prof_number 
          k.unused_class in
        let a8 = TP_Append("KITLIST.2DA",append_to_kitlist,[]) in
        let include_actions = List.map (fun file -> 
          let num = get_next_line_number (file ^ ".2DA" ) in
          let str = Printf.sprintf "%d  %d" num this_kit_number in
          TP_Append(file ^ ".2DA",str,[]) 
        ) include_list in 
        let abbr = Printf.sprintf  "%s     %s" k.kit_name k.tob_abbr in
        let a9 = TP_Append("LUABBR.2DA",abbr,[]) in 
        let a10 = TP_Set_Col("25STWEAP.2DA",
          ("" :: "" :: k.kit_name :: k.tob_start),this_kit_prof_number+1) in 
        let action_list = a1 :: a2 :: a3 :: a4 :: a5 :: a6 :: a7 :: a8 :: 
          a9 :: a10 :: a_e1 :: a_e2 :: include_actions in
        let old_allow_missing = !Load.allow_missing in 
        Load.allow_missing := 
          "LUABBR.2DA" :: "25STWEAP.2DA" :: tp.allow_missing ;
        (* actually do it! *)
        List.iter (process_action tp) action_list ; 
        Load.allow_missing := old_allow_missing ; 
        Var.add_var (k.kit_name) (string_of_int this_kit_number) ;
        log_and_print "Added %s Kit\n" k.kit_name; 
      end
    end 

    | TP_String_Set(s1,str) -> begin
      let i_list = try
        [int_of_string s1]
      with 
        _ -> begin
              try 
                Hashtbl.find_all game.Load.dialog_search s1 
              with _ ->
                log_and_print "ERROR: Cannot find string [%s]\n" s1 ;
                failwith "ERROR: STRING_SET"
             end 
      in 
      List.iter (fun i -> 
        log_and_print "Changing string #%d\n" i ;
        let rec deconstruct ts = match ts with
            Dlg.TLK_Index(_) -> failwith "SET_STRING does not allow #strrefs"
          | Dlg.Local_String(lse) -> lse
          | Dlg.Trans_String(t) -> 
              let new_ts = Hashtbl.find (List.hd !Dc.trans_strings) t in
              deconstruct new_ts 
        in 
        let m,f = Tlk.lse_to_tlk_string (deconstruct str) in 
        Load.set_string game i m f 
      ) i_list 
      end

    | TP_Mkdir(str_l) -> begin
      log_and_print "Creating %d directories\n" (List.length str_l) ;
      List.iter (fun str ->
        try 
          Unix.mkdir str 511 (* 511 = octal 0777 = a+rwx *)
        with _ -> 
          () 
      ) str_l 
      end

    | TP_Compile(dlg_l,tra_l) -> begin
        log_and_print "Compiling %d dialogues/scripts ...\n" 
          (List.length dlg_l) ; 
        Dc.push_copy_trans ();
        begin 
          match !our_lang with
            Some(l) -> List.iter (fun path -> 
                let my_regexp = Str.regexp_string "%s" in
                let tra_file = Str.global_replace 
                  my_regexp l.lang_dir_name path in
                handle_tra_filename tra_file ;
              ) tra_l 
          | _ -> List.iter (fun tra_file -> 
                handle_tra_filename tra_file ; 
              ) tra_l 
        end ;
        let handle_one_d_file d = 
          (* handle AUTO_TRA "solarom/%s" *)
          Dc.push_copy_trans () ;
          List.iter (fun f -> 
            match f,!our_lang with
            Auto_Tra(path),Some(l) -> 
                  let my_regexp = Str.regexp_string "%s" in
                  let tra_file_dir = Str.global_replace 
                    my_regexp l.lang_dir_name path in
                  let d_base,_ = split (Filename.basename d) in
                  let tra_file = tra_file_dir ^ "/" ^ d_base ^ ".TRA" in 
                  handle_tra_filename tra_file ;
          | Auto_Tra(path),None -> 
                  let d_base,_ = split (Filename.basename d) in
                  let tra_file = path ^ "/" ^ d_base ^ ".TRA" in 
                  handle_tra_filename tra_file 
          | _ -> () 
          ) tp.flags ;
          Dc.ok_to_resolve_strings_while_loading := Some(game) ; 
          begin 
          match split (String.uppercase (Filename.basename d)) with
          | _,"BAF" -> compile_baf_filename d
          | _,_ -> handle_d_filename d 
          end ;
          Dc.ok_to_resolve_strings_while_loading := None ; 
          Dc.pop_trans () ; 
        in 
        List.iter (fun filespec -> 
          if is_directory filespec then begin
            try
              let dh = Unix.opendir filespec in 
              while true do
                let base = Unix.readdir dh in 
                let source = filespec ^ "/" ^ base in 
                if not (is_directory source) then
                  handle_one_d_file source 
              done 
            with _ -> () 
          end else handle_one_d_file filespec 
        ) dlg_l ; 
        log_and_print "Processing %d dialogues/scripts ...\n" 
          (List.length dlg_l) ; 
        emit_dlg_files game "override" ;
        Dc.pop_trans () ;
        end

    | TP_Set_Col(file,new_col_list,col_num) -> 
        log_and_print "Setting game text column-wise ...\n" ; 
        let eight,three = split (String.uppercase file) in 
        let buff,loaded_path = Load.load_resource game true eight three in
        if buff = "" then 
          log_or_print "[%s]: empty or does not exist\n" file
        else begin 
          let dest = "override/" ^ file in 
          let buff_as_lines = Str.split (Str.regexp "[\r\n]+") buff in
          if List.length buff_as_lines <> List.length new_col_list then begin
            log_and_print "Cannot set column-wise because there are %d lines in %s but I was only given %d things to append\n" (List.length buff_as_lines) file (List.length new_col_list)  ;
            failwith ("cannot set column-wise to " ^ file)
          end ;
          Stats.time "saving files" (fun () -> 
          let out = open_for_writing dest true in
          List.iter2 (fun orig_line new_col ->
            let line_as_cols = Str.split (Str.regexp "[ \t]+") orig_line in 
            let i = ref 0 in 
            List.iter (fun orig_col -> 
              (if !i = col_num then 
                Printf.fprintf out "%-20s" new_col
              else
                Printf.fprintf out "%-20s" orig_col) ;
              incr i;
            ) line_as_cols ;
            (if (!i <= col_num) then 
                Printf.fprintf out "%-20s" new_col) ; 
            output_string out "\r\n" 
          ) buff_as_lines new_col_list ; 
          close_out out) () ; 
          log_or_print "Set text in [%s] column-wise\n" file 
        end 

    | TP_Append_Col(file,src_list,con_l) ->
        log_and_print "Appending game text column-wise ...\n" ; 
        let eight,three = split (String.uppercase file) in 
        let buff,loaded_path = Load.load_resource game true eight three in
        if buff = "" then 
          log_or_print "[%s]: empty or does not exist\n" file
        else begin 
        let okay = List.fold_left (fun acc elt -> acc &&
          match elt with 
            TP_Contains(s) -> begin
              let my_regexp = Str.regexp_string s in
              try let _ = Str.search_forward my_regexp buff 0 in
                log_only "May append cols to [%s] because it DOES contain [%s]\n" file s  ; 
                true
              with _ -> 
                log_only "Not appending cols to [%s] because it does NOT contain [%s]\n" file s ; 
                false
              end
          | TP_NotContains(s) -> begin
              let my_regexp = Str.regexp_string s in
              try let _ = Str.search_forward my_regexp buff 0 in 
                log_only "Not appending cols to [%s] because it DOES contains [%s]\n" file s ; 
                false
              with _ -> 
                log_only "May append cols to [%s] because it does NOT contain [%s]\n" file s ; 
                true
              end
          | TP_IfSizeIs(size) -> String.length buff = size 
          | TP_Eval(pe) -> 
                let v = eval_pe pe in 
                if v = Int32.zero then begin
                  log_only "Not appending cols to [%s] because condition evaluates to %ld\n" file v ; false
                end else begin 
                  log_only "May append cols to [%s] because condition evaluates to %ld\n" file v ; true
                end 
        ) true con_l in 
        if okay then begin (* do the append *) 
          let dest = "override/" ^ file in 
          let buff_as_lines = Str.split (Str.regexp "[\r\n]+") buff in
          if List.length buff_as_lines <> List.length src_list then begin
            log_and_print "Cannot append column-wise because there are %d lines in %s but I was only given %d things to append\n" (List.length buff_as_lines) file (List.length src_list)  ;
            failwith ("cannot append column-wise to " ^ file)
          end ;
          Stats.time "saving files" (fun () -> 
          let out = open_for_writing dest true in
          List.iter2 (fun orig app ->
            output_string out orig ;
            output_string out " " ;
            output_string out app ;
            output_string out "\r\n" 
          ) buff_as_lines src_list ; 
          close_out out) () ; 
          log_or_print "Appended text to [%s] column-wise\n" file 
        end 
        end

    | TP_Append(file,src,con_l) ->
        log_and_print "Appending game text ...\n" ; 
        let eight,three = split (String.uppercase file) in 
        let buff,loaded_path = Load.load_resource game true eight three in
        let okay = List.fold_left (fun acc elt -> acc &&
          match elt with 
            TP_Contains(s) -> begin
              let my_regexp = Str.regexp_string s in
              try let _ = Str.search_forward my_regexp buff 0 in
                log_only "May append [%.10s...] to [%s] because it DOES contain [%s]\n" src file s  ; 
                true
              with _ -> 
                log_only "Not appending [%.10s...] to [%s] because it does NOT contain [%s]\n" src file s ; 
                false
              end
          | TP_NotContains(s) -> begin
              let my_regexp = Str.regexp_string s in
              try let _ = Str.search_forward my_regexp buff 0 in 
                log_only "Not appending [%.10s...] to [%s] because it DOES contains [%s]\n" src file s ; 
                false
              with _ -> 
                log_only "May append [%.10s...] to [%s] because it does NOT contain [%s]\n" src file s ; 
                true
              end
          | TP_IfSizeIs(size) -> String.length buff = size 
          | TP_Eval(pe) -> 
                let v = eval_pe pe in 
                if v = Int32.zero then begin
                  log_only "Not appending [%.10s] to [%s] because condition evaluates to %ld\n" src file v ; false
                end else begin 
                  log_only "May appending [%.10s] to [%s] because condition evaluates to %ld\n" src file v ; true
                end 
        ) true con_l in 
        if okay then begin (* do the append *) 
          let dest = "override/" ^ file in 
          Stats.time "saving files" (fun () -> 
          let out = open_for_writing dest true in
          let nice_newlines = Str.global_replace
            (Str.regexp "[\r\n]+") "\r\n" (buff ^ "\r\n") in
          output_string out nice_newlines ; 
          (* output_string out buff ;
          output_string out "\r\n" ; *)
          output_string out src ;
          output_string out "\r\n" ;
          close_out out) () ; 
          log_or_print "Appended text to [%s]\n" file 
        end 

    | TP_Extend_Top(use_reg,dest,src,pl,tra_l) 
    | TP_Extend_Bottom(use_reg,dest,src,pl,tra_l) -> begin
        log_and_print "Extending game scripts ...\n" ; 
        let dlist = 
          if use_reg = false then
            [dest]
          else begin
            let files_in_chitin = Key.list_of_key_resources game.Load.key in 
            let regexp = Str.regexp_case_fold dest in 
            let matches = ref [] in 
            List.iter (fun possible ->
              if Str.string_match regexp possible 0 then begin
                matches := (possible) :: !matches
              end 
            ) files_in_chitin ;
            if (!matches = []) then
              [dest]
            else
              !matches
          end 
        in 
        Dc.push_copy_trans () ;
        begin 
          match !our_lang with
            Some(l) -> List.iter (fun path -> 
                let my_regexp = Str.regexp_string "%s" in
                let tra_file = Str.global_replace 
                  my_regexp l.lang_dir_name path in
                handle_tra_filename tra_file ;
              ) tra_l 
          | _ -> List.iter (fun tra_file -> 
                handle_tra_filename tra_file ; 
              ) tra_l 
        end ;
        let src_buff = load_file src in 
        List.iter (fun p -> process_patch1 game src_buff p) pl ; 
        let src_buff = List.fold_left (fun acc elt -> 
              try process_patch2 game acc elt
              with e -> 
                log_and_print "ERROR: [%s] -> [%s] Patching Failed (EXTEND_TOP/BOTTOM)\n" 
                  src dest ; raise e) 
            src_buff pl 
        in 
        Dc.ok_to_resolve_strings_while_loading := Some(game) ; 
        let src_script = handle_script_buffer src src_buff in
        Dc.ok_to_resolve_strings_while_loading := None ; 
        List.iter (fun dest -> 
          let eight,three = split (String.uppercase dest) in 
          let dest_buff,loaded_path = 
            Load.load_resource game true eight three in
          let dest_script = handle_script_buffer dest dest_buff in 
          let destpath = "override/" ^ dest in 
          Stats.time "saving files" (fun () -> 
          let out = open_for_writing destpath true in
          Bcs.save_bcs game (Bcs.Save_BCS_OC(out)) (match a with 
            TP_Extend_Top(_,_,_,_,_) -> src_script @ dest_script
                                 | _ -> dest_script @ src_script) ;
          close_out out) () ; 
          log_or_print "Extended script [%s] with [%s]\n" dest src
        ) dlist ;
        Dc.pop_trans () ; 
        end

    | TP_At_Interactive_Exit(str) ->  
        if !interactive then process_action tp (TP_At_Exit(str))
    | TP_At_Interactive_Uninstall(str) ->
        if !interactive then process_action tp (TP_At_Uninstall(str))

    | TP_At_Exit(str) ->  
        let str = Arch.handle_view_command str in 
        if List.mem str !execute_at_exit then
          ()
        else
          execute_at_exit := str :: !execute_at_exit 

    | TP_At_Uninstall(str) -> () 
  in 
  let lang_init () = 
    begin 
      match !our_lang with
        None -> () 
      | Some(l) -> 
          log_and_print "Using Language [%s]\n" l.lang_name ;
          log_or_print "[%s] has %d top-level TRA files\n"
            l.lang_name (List.length l.lang_tra_files) ;
          List.iter handle_tra_filename l.lang_tra_files
    end 
  in 
  lang_init () ; 

  let num = List.length tp.modules in 

  let any_already_installed = ref false in 
  let any_not_yet_installed = ref false in 

  for i = 0 to num-1 do 
    if already_installed this_tp2_filename i then
      any_already_installed := true
    else begin
      let m = List.nth tp.modules i in
      let deprecated = List.exists (fun f -> match f with 
        | TPM_Deprecated(warn) -> true
        | _ -> false) m.mod_flags in 
      if not deprecated then
        any_not_yet_installed := true 
    end 
  done ; 

  let module_defaults = Array.init num (fun i -> 
      if !always_yes then
        TP_Install
      else if !always_uninstall && (already_installed this_tp2_filename i) then
        TP_Uninstall
      else if !always_uninstall then
        TP_Skip
      else 
        TP_Ask) 
  in

  let handle_letter tp answer can_uninstall package_name m finished i =
    match answer with 
    | "Q" -> begin
        for i = 0 to num - 1 do
          module_defaults.(i) <- TP_Skip ; 
          finished := true 
        done 
        end 
    | "Y" | "R" -> begin
        if can_uninstall then begin 
          try 
            log_and_print 
            "\nRemoving old installation of [%s] first ...\n" package_name ;
            (if not (uninstall handle_tp2_filename this_tp2_filename i !interactive) then failwith "uninstallation error"); 
            log_and_print "\nSUCCESSFULLY REMOVED OLD [%s]\n\n" package_name ;
          with e ->
            log_and_print "WARNING: unable to uninstall: %s\n" 
              (Printexc.to_string e)
        end ; 
        set_backup_dir tp.backup i ; 
        log_and_print "\nInstalling [%s]\n" package_name ;
        (try 
          List.iter (fun flag -> match flag with
            Always(al) -> List.iter (process_action tp) al
          | _ -> () 
          ) tp.flags ;
          List.iter (process_action tp) m.mod_parts ;
        with e -> begin
          log_and_print "\nERROR Installing [%s], rolling back to previous state\n" package_name ; 
          Dc.clear_state () ; 
          (match !backup_list_chn with
            Some(chn) -> close_out chn ; backup_list_chn := None
          | None -> ()) ;
          uninstall_tp2_component tp this_tp2_filename i false ;
          print_log () ; 
          raise e
        end ); 
        log_and_print "\n\n" ; 
        strings_to_print_at_exit := 
          (Printf.sprintf "\nSUCCESSFULLY INSTALLED [%s]\n" package_name) :: 
          !strings_to_print_at_exit ;
        (* add this successful install to the log! *)
        the_log := !the_log @ 
          [ ((String.uppercase this_tp2_filename),!our_lang_index,i,Some(package_name),Installed) ] ; 
        finished := true 
      end
    | "N" -> 
        log_and_print "\nSkipping [%s]\n" package_name ;
        finished := true 
    | "U" -> 
        log_and_print "\nRemoving [%s]\n" package_name ;
        (if not (uninstall handle_tp2_filename this_tp2_filename i !interactive ) then failwith "uninstallation error" );
        log_and_print "\n\nSUCCESSFULLY REMOVED [%s]\n\n" package_name ;
        finished := true 
    | _ -> () 
  in 

  (* for big mods, ask about things in general first *) 
  if num > 4 && not !always_yes && not !always_uninstall &&
     not (List.exists (fun a -> a = Ask_Every_Component) tp.flags)
  then begin
      log_and_print "\nThis mod has %d distinct optional components.\nTo save time, you can choose what to do with them at a high level rather\nthan being asked about each one.\n" num ;
      let finished = ref false in

      if !any_not_yet_installed then 
      while not !finished do 
        finished := true ;
        log_and_print "\nWhat should be done with all components that are NOT YET installed?\n";
        log_and_print "[I]nstall them, [S]kip them, [A]sk about each one? " ;
        match String.uppercase(read_line ()) with
        | "R"
        | "I" -> 
          for i = 0 to num - 1 do 
            if not (already_installed this_tp2_filename i) then
              module_defaults.(i) <- TP_Install
          done 
        | "S" ->
          for i = 0 to num - 1 do 
            if not (already_installed this_tp2_filename i) then
              module_defaults.(i) <- TP_Skip
          done 
        | "A" -> ()
        | _ -> finished := false 
      done ;

      finished := false ; 
      if !any_already_installed then 
      while not !finished do
        finished := true ;
        log_and_print "\nWhat should be done with all components that are ALREADY installed?\n";
        log_and_print "[R]e-install them, [U]ninstall them, [S]kip them, [A]sk about each one? " ;
        match String.uppercase(read_line ()) with
        | "I" 
        | "R" -> 
          for i = 0 to num - 1 do 
            if (already_installed this_tp2_filename i) then
              module_defaults.(i) <- TP_Install
          done 
        | "S" ->
          for i = 0 to num - 1 do 
            if (already_installed this_tp2_filename i) then
              module_defaults.(i) <- TP_Skip
          done 
        | "U" ->
          for i = 0 to num - 1 do 
            if (already_installed this_tp2_filename i) then
              module_defaults.(i) <- TP_Uninstall
          done 
        | "A" -> ()
        | _ -> finished := false 
      done ;
  end ; 


  let original_menu_style () = 
    let current = ref (-1) in 
    List.iter (fun m ->
      incr current ; 
      let def = ref (module_defaults.(!current)) in 
      let package_name = Dc.single_string_of_tlk_string game m.mod_name in 
      let can_uninstall = already_installed this_tp2_filename !current in 
      print_log () ; 
      List.iter (fun f -> match f with
      | TPM_Deprecated(warn) -> 
        begin 
          if can_uninstall then begin
            let warn = Dc.single_string_of_tlk_string game warn in 
            log_and_print "\nNOTE: [%s] is deprecated. Uninstalling!\n\t%s\n" 
              package_name warn ;
            def := TP_Uninstall ; 
          end else begin
            def := TP_Skip ; 
          end 
        end 
      | TPM_RequireComponent(s,i,warn) -> 
        begin
          if already_installed s i then
            () (* good! *) 
          else if (!def <> TP_Skip) && 
                  (!def <> TP_Uninstall) then begin
            let warn = Dc.single_string_of_tlk_string game warn in 
            log_and_print "\nSKIPPING: [%s]\n\t%s\n" package_name warn ;
            def := TP_Skip ; 
          end 
        end 
      ) m.mod_flags ; 
      let finished = ref false in 
      let handle_error e = 
        log_and_print "ERROR: %s\n" (Printexc.to_string e) ;
        Dc.clear_state () ; 
        (if (!log_file <> "") then 
        log_and_print "PLEASE email the file %s to %s\n" !log_file tp.author) ;
        if !always_yes then begin
          log_and_print "Automatically Skipping [%s] because of error.\n"
            package_name ; 
          finished := true 
        end ; 
        lang_init () 
      in 
      Dc.clear_state () ; 
      match !def with
      | TP_Install ->
        begin try 
          handle_letter tp "Y" can_uninstall package_name m finished !current ;
          with e -> handle_error e  
        end 
      | TP_Uninstall ->
        begin try
          handle_letter tp "U" can_uninstall package_name m finished !current ;
          with e -> handle_error e
        end 
      | TP_Skip -> ()
      | TP_Ask -> 
        while not !finished do try
          let can_uninstall = already_installed this_tp2_filename !current in 
          if can_uninstall then 
            log_and_print "\nInstall Component [%s]\n[R]e-Install, [N]o Change or [U]ninstall or [Q]uit? "  package_name 
          else log_and_print "\nInstall Component [%s]\n[Y]es or [N]o or [Q]uit? "  package_name ;
          begin
          let answer = String.uppercase(read_line ()) in 
          handle_letter tp answer can_uninstall package_name m
                      finished !current ;
          Dc.clear_state () ; 
          end
        with e -> handle_error e
        done 
    ) tp.modules
  in 

  original_menu_style () ; 

  interactive := false ; 

  (* now we must handle every temporarily-uninstalled mods *)
  let re_installed = ref [] in 
  let rec process lst = match lst with
    [] -> []
  | (_,_,_,_,Installed) as head :: tl -> head :: (process tl)
  | (_,_,_,_,Permanently_Uninstalled) as head :: tl -> head :: (process tl)
  | (a,b,c,sopt,Temporarily_Uninstalled) as head :: tl -> 
    begin 
    try 
      (* we must re-install it! *)
      begin
        log_and_print "\nRe-Installing [%s] component %d %s\n" a c 
          (str_of_str_opt sopt); 
        let tp_file = a in 
        let tp2 = handle_tp2_filename tp_file in 
        (* load their chosen language *)
        Dc.clear_state () ;
        Dc.push_trans ();
        (try
          let l = List.nth tp2.languages b in
          our_lang := Some(l) ;
          our_lang_index := b ; 
          log_and_print "Re-Installing Using Language [%s]\n" l.lang_name ;
          List.iter handle_tra_filename l.lang_tra_files
        with _ -> 
          our_lang := None ; 
          our_lang_index := 0 ;
          () ) ;
        let m = List.nth tp2.modules c in 
        let package_name = Dc.single_string_of_tlk_string game m.mod_name in 
        let fails_requirements = List.exists (fun f -> match f with 
          | TPM_RequireComponent(s,i,warn) -> 
            begin
              if already_installed s i then
                false
              else begin
                log_and_print "\n[%s] component %d %s fails component requirements, *not* Re-Installing.\n" a c (str_of_str_opt sopt); 
                true
              end 
            end
          | _ -> false) m.mod_flags 
        in 
        begin 
        if fails_requirements then begin 
          handle_letter tp2 "U" false package_name m (ref false) c ;
        end else begin 
          handle_letter tp2 "Y" false package_name m (ref false) c ;
        end  
        end ; 
        Dc.clear_state () ; 
        Dc.pop_trans ();
      end ;
      re_installed := !re_installed @ [(a,b,c,sopt,Installed)] ;
      (process tl)  
    with e ->
      log_and_print "ERROR Re-Installing [%s] component %d %s\nTry to re-install it manually.\n%s\n" a c (str_of_str_opt sopt) (Printexc.to_string e) ;
      (a,b,c,sopt,Permanently_Uninstalled) :: (process tl)
    end 
  in 
  let result = (process !the_log) @ !re_installed in 

  the_log := result ; 

  save_log () ; 

  Load.allow_missing := []
end
